Package: reproducible
Type: Package
Title: Enhance Reproducibility of R Code
Description: A collection of high-level, machine- and OS-independent tools
    for making reproducible and reusable content in R.
    The two workhorse functions are 'Cache()' and 'prepInputs()'.
    'Cache()' allows for nested caching, is robust to environments and objects
    with environments (like functions), and deals with some classes of 
    file-backed R objects e.g., from 'terra' and 'raster' packages. 
    Both functions have been developed to be foundational components of data
    retrieval and processing in continuous workflow situations. In both functions,
    efforts are made to make the first and subsequent calls of functions have 
    the same result, but faster at subsequent times by way of checksums
    and digesting. Several features are still under development, including
    cloud storage of cached objects allowing for sharing between users. Several
    advanced options are available, see '?reproducibleOptions()'.
SystemRequirements: 'unrar' (Linux/macOS) or '7-Zip' (Windows) to work
        with '.rar' files.
URL: https://reproducible.predictiveecology.org,
        https://github.com/PredictiveEcology/reproducible
Date: 2026-05-17
Version: 3.1.1
Authors@R: 
    c(person(given = "Eliot J B",
             family = "McIntire",
             role = c("aut", "cre"),
             email = "eliot.mcintire@canada.ca",
             comment = c(ORCID = "0000-0002-6914-8316")),
      person(given = "Alex M",
             family = "Chubaty",
             role = c("aut"),
             email = "achubaty@for-cast.ca",
             comment = c(ORCID = "0000-0001-7146-8135")),
      person(given = "Tati",
             family = "Micheletti",
             role = "ctb",
             email = "tati.micheletti@gmail.com",
             comment = c(ORCID = "0000-0003-4838-8342")),
      person(given = "Ceres",
             family = "Barros",
             role = "ctb",
             email = "ceres.barros@ubc.ca",
             comment = c(ORCID = "0000-0003-4036-977X")),
      person(given = "Ian",
             family = "Eddy",
             role = "ctb",
             email = "ian.eddy@nrcan-rncan.gc.ca",
             comment = c(ORCID = "0000-0001-7397-2116")),
      person(given = "His Majesty the King in Right of Canada, as represented by the Minister of Natural Resources Canada",
             role = "cph"))
Depends: R (>= 4.3)
Imports: cli, data.table (>= 1.10.4), digest, filelock (>= 1.0.3),
        fpCompare, fs, lobstr, methods, stats, tools, utils
Suggests: archive, covr, curl (>= 7.0.0), DBI, future, geodata, glue,
        googledrive, httr, httr2 (>= 1.2.1), knitr, parallel,
        parallelly, qs2, raster (>= 3.5-15), RCurl (>= 1.95-4.8),
        rlang, rmarkdown, RSQLite, R.utils, rvest, sf, sp (>= 1.4-2),
        terra (>= 1.7-20), testthat, withr
Encoding: UTF-8
Language: en-CA
License: GPL-3
VignetteBuilder: knitr, rmarkdown
BugReports: https://github.com/PredictiveEcology/reproducible/issues
ByteCompile: yes
Config/roxygen2/version: 8.0.0
Collate: 'DBI.R' 'messages.R' 'GPT2.R' 'cache-helpers.R'
        'cache-internals.R' 'robustDigest.R' 'cache.R' 'cacheGeo.R'
        'checksums.R' 'cloud.R' 'convertPaths.R' 'copy.R' 'download.R'
        'downloadTileAndUpload.R' 'exportedMethods.R' 'gis.R'
        'helpers.R' 'listNamed.R' 'objectSize.R' 'options.R'
        'packages.R' 'paths.R' 'pipe.R' 'postProcess.R'
        'postProcessTo.R' 'preProcess.R' 'prepInputs.R'
        'prepInputsCOG.R' 'reproducible-deprecated.R'
        'reproducible-package.R' 'search.R' 'showCacheEtc.R'
        'spatialObjects-class.R' 'terra-migration.R' 'zzz.R'
NeedsCompilation: no
Packaged: 2026-05-18 04:44:29 UTC; emcintir
Author: Eliot J B McIntire [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6914-8316>),
  Alex M Chubaty [aut] (ORCID: <https://orcid.org/0000-0001-7146-8135>),
  Tati Micheletti [ctb] (ORCID: <https://orcid.org/0000-0003-4838-8342>),
  Ceres Barros [ctb] (ORCID: <https://orcid.org/0000-0003-4036-977X>),
  Ian Eddy [ctb] (ORCID: <https://orcid.org/0000-0001-7397-2116>),
  His Majesty the King in Right of Canada, as represented by the Minister
    of Natural Resources Canada [cph]
Maintainer: Eliot J B McIntire <eliot.mcintire@canada.ca>
Repository: CRAN
Date/Publication: 2026-05-18 07:30:08 UTC
Built: R 4.6.0; ; 2026-05-18 10:04:20 UTC; unix
