# Generated by roxygen2: do not edit by hand

S3method(.unwrap,PackedSpatExtent2)
S3method(.unwrap,PackedSpatVector)
S3method(.unwrap,PackedSpatVector2)
S3method(.unwrap,data.table)
S3method(.unwrap,default)
S3method(.unwrap,environment)
S3method(.unwrap,list)
S3method(.wrap,default)
S3method(.wrap,environment)
S3method(.wrap,list)
S3method(asPath,character)
S3method(asPath,null)
S3method(assessDataType,default)
S3method(cropInputs,default)
S3method(fixErrors,default)
S3method(makeMemoisable,data.table)
S3method(makeMemoisable,default)
S3method(maskInputs,default)
S3method(objSize,"function")
S3method(objSize,Path)
S3method(objSize,default)
S3method(objSize,environment)
S3method(objSize,list)
S3method(postProcess,default)
S3method(postProcess,list)
S3method(projectInputs,default)
S3method(unmakeMemoisable,default)
S3method(updateFilenameSlots,default)
S3method(updateFilenameSlots,environment)
S3method(updateFilenameSlots,list)
S3method(usesPointer,default)
S3method(usesPointer,environment)
S3method(usesPointer,list)
S3method(writeOutputs,default)
export(.addChangedAttr)
export(.addTagsRepo)
export(.addTagsToOutput)
export(.cacheMessage)
export(.cacheMessageObjectToRetrieve)
export(.checkCacheRepo)
export(.file.move)
export(.formalsNotInCurrentDots)
export(.grepSysCalls)
export(.isCRSany)
export(.isGridded)
export(.isMemoised)
export(.isSF)
export(.isSpat)
export(.isSpatialAny)
export(.isVector)
export(.objSizeWithTry)
export(.orderDotsUnderscoreFirst)
export(.preDigestByClass)
export(.prefix)
export(.prepareOutput)
export(.requireNamespace)
export(.robustDigest)
export(.setSubAttrInList)
export(.sortDotsUnderscoreFirst)
export(.suffix)
export(.tagsByClass)
export(.unwrap)
export(.updateTagsRepo)
export(.whereInStack)
export(.wrap)
export(Cache)
export(CacheDBFile)
export(CacheDBTableName)
export(CacheDigest)
export(CacheGeo)
export(CacheIsACache)
export(CacheStorageDir)
export(CacheStoredFile)
export(CacheV2)
export(Checksums)
export(Copy)
export(Filenames)
export(asPath)
export(assessDataType)
export(basename2)
export(cacheId)
export(cc)
export(checkAndMakeCloudFolderID)
export(checkPath)
export(clearCache)
export(compareNA)
export(convertPaths)
export(convertRasterPaths)
export(copyFile)
export(createCache)
export(cropInputs)
export(cropTo)
export(dataType2)
export(detectActiveCores)
export(downloadFile)
export(extractFromCache)
export(fixErrors)
export(fixErrorsIn)
export(gdalMask)
export(gdalProject)
export(gdalResample)
export(getRelative)
export(internetExists)
export(isUpdated)
export(linkOrCopy)
export(listNamed)
export(loadFile)
export(loadFromCache)
export(makeMemoisable)
export(makeRelative)
export(maskInputs)
export(maskTo)
export(maxFn)
export(messageCache)
export(messageColoured)
export(messageDF)
export(messagePreProcess)
export(messagePrepInputs)
export(messageQuestion)
export(minFn)
export(movedCache)
export(nlayers2)
export(normPath)
export(normPathRel)
export(numCoresToUse)
export(objSize)
export(objSizeSession)
export(paddedFloatToChar)
export(postProcess)
export(postProcessTerra)
export(postProcessTo)
export(preProcess)
export(preProcessParams)
export(prepInputs)
export(prepInputsCOG)
export(prepInputsWithTiles)
export(prepopulateCacheAsync)
export(projectInputs)
export(projectTo)
export(purgeChecksums)
export(rasterRead)
export(remapFilenames)
export(reproducibleOptions)
export(retry)
export(rmFromCache)
export(searchFull)
export(searchFullEx)
export(set.randomseed)
export(showCache)
export(studyAreaName)
export(tempdir2)
export(tempfile2)
export(unmakeMemoisable)
export(updateFilenameSlots)
export(urlExists)
export(usesPointer)
export(values2)
export(writeFuture)
export(writeOutputs)
export(writeTo)
exportClasses(Path)
exportMethods(.addChangedAttr)
exportMethods(.addTagsToOutput)
exportMethods(.cacheMessage)
exportMethods(.checkCacheRepo)
exportMethods(.preDigestByClass)
exportMethods(.prepareOutput)
exportMethods(.robustDigest)
exportMethods(.tagsByClass)
exportMethods(Filenames)
exportMethods(checkPath)
exportMethods(clearCache)
exportMethods(keepCache)
exportMethods(mergeCache)
exportMethods(normPath)
exportMethods(showCache)
exportMethods(studyAreaName)
import(methods)
importFrom(data.table,":=")
importFrom(data.table,.N)
importFrom(data.table,.SD)
importFrom(data.table,as.data.table)
importFrom(data.table,copy)
importFrom(data.table,data.table)
importFrom(data.table,fread)
importFrom(data.table,is.data.table)
importFrom(data.table,rbindlist)
importFrom(data.table,set)
importFrom(data.table,setDF)
importFrom(data.table,setDT)
importFrom(data.table,setcolorder)
importFrom(data.table,setindex)
importFrom(data.table,setkeyv)
importFrom(data.table,setnames)
importFrom(data.table,setorderv)
importFrom(digest,digest)
importFrom(fpCompare,"%==%")
importFrom(fs,is_absolute_path)
importFrom(fs,path_abs)
importFrom(fs,path_expand)
importFrom(fs,path_join)
importFrom(fs,path_norm)
importFrom(lobstr,obj_size)
importFrom(methods,"slot<-")
importFrom(methods,.S4methods)
importFrom(methods,formalArgs)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,selectMethod)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(stats,na.omit)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(utils,capture.output)
importFrom(utils,download.file)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,isS3stdGeneric)
importFrom(utils,methods)
importFrom(utils,modifyList)
importFrom(utils,object.size)
importFrom(utils,packageVersion)
importFrom(utils,read.table)
importFrom(utils,tail)
importFrom(utils,untar)
importFrom(utils,unzip)
importFrom(utils,write.table)
