################################################################################
#' The `spatialClasses` class
#'
#' This class is defunct.
#'
#' @aliases spatialClasses
#' @name spatialClasses-class
#' @rdname spatialClasses-class
#' @author Eliot McIntire
