% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/listNamed.R
\name{listNamed}
\alias{listNamed}
\title{Create a list with names from object names}
\usage{
listNamed(...)
}
\arguments{
\item{...}{Any elements to add to a list, as in \code{base::list}}
}
\description{
This is a convenience wrapper around \verb{a <- 1; newList <- list(a); names(newList) <- "a"}.
}
\details{
This will return a named list, where names are the object names, captured
internally in the function and assigned to the list. If a user manually supplies
names, these will be kept (i.e., not overwritten by the object name).
}
\examples{
a <- 1
b <- 2
d <- 3
(newList <- listNamed(a, b, dManual = d)) # "dManual" name kept


}
