% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{.whereInStack}
\alias{.whereInStack}
\title{Search for objects in the call stack}
\usage{
.whereInStack(obj, startingEnv = parent.frame())
}
\arguments{
\item{obj}{Character string. The object name to search.}

\item{startingEnv}{An environment to start searching in.}
}
\value{
The environment in which the object exists. It will return the
first environment it finds, searching outwards from where the function is used.
}
\description{
Normally, this is only used in special, advanced uses. The standard approach
to getting an object from an environment in the call stack is to explicitly
pass it into the function.
}
