% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportedMethods.R
\name{.grepSysCalls}
\alias{.grepSysCalls}
\title{Grep system calls}
\usage{
.grepSysCalls(sysCalls, pattern)
}
\arguments{
\item{sysCalls}{The return from \code{sys.calls()}}

\item{pattern}{Character, passed to grep}
}
\value{
Numeric vector, equivalent to return from \code{grep(sys.calls(), pattern = "test")},
but faster if \code{sys.calls()} is very big.
}
\description{
A faster way of grepping the system call stack than just
\code{grep(sys.calls(), pattern = "test")}
}
\keyword{internal}
