% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{purgeChecksums}
\alias{purgeChecksums}
\title{Purge the checksums of a single file}
\usage{
purgeChecksums(checksumFile, fileToRemove)
}
\arguments{
\item{checksumFile}{A character string indicating the absolute path to the \code{CHECKSUMS.txt}
file.}

\item{fileToRemove}{The filename to remove from the \code{checksumFile}}
}
\value{
NULL. Run for its side effect, namely, and file removed from the \file{CHECKSUMS.txt}
file.
}
\description{
This is a manual way of achieving \code{prepInputs(..., purge = 7)}, useful in cases
where \code{prepInputs} is not called directly by the user, so it would be difficult
to set \code{purge = 7}.
}
