% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/copy.R
\name{.file.move}
\alias{.file.move}
\title{Move a file to a new location -- Defunct -- use \code{hardLinkOrCopy}}
\usage{
.file.move(from, to, overwrite = FALSE)
}
\arguments{
\item{from, to}{character vectors, containing file names or paths.}

\item{overwrite}{logical indicating whether to overwrite destination file if it exists.}
}
\value{
Logical indicating whether operation succeeded.
}
\description{
This will first try to \code{file.rename}, and if that fails, then it will
\code{file.copy} then \code{file.remove}.
}
