% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache-helpers.R
\name{.debugCache}
\alias{.debugCache}
\title{Attach debug info to return for Cache}
\usage{
.debugCache(obj, preDigest, ..., fullCall)
}
\arguments{
\item{obj}{An arbitrary R object.}

\item{preDigest}{A list of hashes.}

\item{...}{Dots passed from \code{Cache}}

\item{fullCall}{The original call to \code{Cache}}
}
\value{
The same object as \code{obj}, but with 2 attributes set.
}
\description{
Internal use only. Attaches an attribute to the output, usable for
debugging the Cache.
}
\author{
Eliot McIntire
}
