% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{cacheId}
\alias{cacheId}
\title{Extract the cache id of an object}
\usage{
cacheId(obj)
}
\arguments{
\item{obj}{Any R object}
}
\value{
The \code{cacheId} if this was part of a \code{Cache} call. Otherwise \code{NULL}
}
\description{
Any object that was returned from the Cache or was calculated as part of a
Cache call will have an attribute, \code{tags} and an entry with \verb{cacheId:} prefix.
This is a lightweight helper to extract that \code{cacheId}.
}
