% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{guessAtTarget}
\alias{guessAtTarget}
\alias{.guessAtTargetAndFun}
\title{Try to pick a file to load}
\usage{
.guessAtTargetAndFun(
  targetFilePath,
  destinationPath = getOption("reproducible.destinationPath", "."),
  filesExtracted,
  fun = NULL,
  verbose = getOption("reproducible.verbose", 1)
)
}
\arguments{
\item{destinationPath}{Full path of the directory where the target file should be}

\item{filesExtracted}{A character vector of all files that have been extracted (e.g.,
from an archive)}
}
\description{
Try to pick a file to load
}
\keyword{internal}
