% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{.setSubAttrInList}
\alias{.setSubAttrInList}
\title{Set subattributes within a list by reference}
\usage{
.setSubAttrInList(object, attr, subAttr, value)
}
\arguments{
\item{object}{An arbitrary object}

\item{attr}{The attribute name (that is a list object) to change}

\item{subAttr}{The list element name to change}

\item{value}{The new value}
}
\value{
This sets or updates the \code{subAttr} element of a list that is located at
\code{attr(object, attr)}, with the \code{value}. This, therefore, updates a sub-element
of a list attribute and returns that same object with the updated attribute.
}
\description{
Sets only a single element within a list attribute.
}
