% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache-helpers.R
\name{isUpdated}
\alias{isUpdated}
\title{Has a cached object has been updated?}
\usage{
isUpdated(x)
}
\arguments{
\item{x}{cached object}
}
\value{
logical
}
\description{
Has a cached object has been updated?
}
