% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/messages.R, R/zzz.R
\name{.message}
\alias{.message}
\alias{.pkgEnv}
\title{The \code{reproducible} package environments}
\usage{
.message

.pkgEnv
}
\description{
Environment used internally to store internal package objects and methods.
}
\details{
\itemize{
\item \code{.message} is specifically for messages and message-generating functions;
\item \code{.pkgEnv} is for general use within the package;
\item \code{.reproEnv} is used for \code{Cache}-related objects;
}
}
\keyword{internal}
