% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportedMethods.R
\name{remapFilenames}
\alias{remapFilenames}
\title{Remap file names}
\usage{
remapFilenames(obj, tags, cachePath = getOption("reproducible.cachePath"), ...)
}
\arguments{
\item{obj}{(optional) object whose file path metadata will be remapped}

\item{tags}{cache tags \code{data.table} object}

\item{cachePath}{character string specifying the path to the cache directory or \code{NULL}}

\item{...}{Additional path arguments, passed to \code{absoluteBase()} and \code{modifyListPaths()}}
}
\description{
Update file path metadata for file-backed objects (e.g., \code{SpatRasters}).
Useful when moving saved objects between projects or machines.
}
