% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{isInteractive}
\alias{isInteractive}
\title{Alternative to \code{interactive()} for unit testing}
\usage{
isInteractive()
}
\description{
This is a suggestion from
\url{https://github.com/MangoTheCat/blog-with-mock/blob/master/Blogpost1.Rmd}
as a way to test interactive code in unit tests. Basically, in the unit tests,
we use \code{testthat::with_mock}, and inside that we redefine \code{isInteractive}
just for the test. In all other times, this returns the same things as
\code{interactive()}.
}
\keyword{internal}
