% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{archiveExtractBinary}
\alias{archiveExtractBinary}
\alias{.archiveExtractBinary}
\title{Tests if unrar or 7zip exist}
\usage{
.archiveExtractBinary(verbose = getOption("reproducible.verbose", 1))
}
\value{
unrar or 7zip path if exist, or NULL
}
\description{
Tests if unrar or 7zip exist
}
\author{
Tati Micheletti
}
\keyword{internal}
