% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{unrarPath}
\alias{unrarPath}
\alias{.systemArchivePath}
\title{The known path for unrar or 7z}
\usage{
.systemArchivePath
}
\description{
The known path for unrar or 7z
}
