% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DBI.R
\name{usesPointer}
\alias{usesPointer}
\title{Does an object use a pointer?}
\usage{
usesPointer(x)
}
\arguments{
\item{x}{an object}
}
\value{
logical
}
\description{
Does an object use a pointer?
}
