% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{testForArchiveExtract}
\alias{testForArchiveExtract}
\alias{.testForArchiveExtract}
\title{Returns \code{unrar} path and creates a shortcut as \code{.systemArchivePath}.
Was not incorporated in previous function so it can be used in the tests.}
\usage{
.testForArchiveExtract(archive = "")
}
\value{
path to \code{unrar} or \code{7-zip} if it exists, and assign it to \code{.systemArchivePath}.
Advise the user if not found..
}
\description{
Returns \code{unrar} path and creates a shortcut as \code{.systemArchivePath}.
Was not incorporated in previous function so it can be used in the tests.
}
\author{
Tati Micheletti
}
\keyword{internal}
