% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{.listFilesInArchive}
\alias{.listFilesInArchive}
\title{List files in either a \code{.zip} or or \code{.tar} file}
\usage{
.listFilesInArchive(archive, dontUse = NULL)
}
\arguments{
\item{archive}{A character string of a single file name to list files in.}

\item{dontUse}{A character string or vector of unzipping tool(s) to not use,
e.g., one or more of: \code{c("archive", "unzip", "7z")}}
}
\value{
A character string of all files in the archive.
}
\description{
Makes the outputs from\verb{.tar``.zip} the same, which they aren't by default.
}
\keyword{internal}
