% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{internetExists}
\alias{internetExists}
\alias{urlExists}
\title{Checks for existed of a url or the internet using \url{https://CRAN.R-project.org}}
\usage{
internetExists()

urlExists(url)
}
\arguments{
\item{url}{A url of the form \verb{https://...} to test for existence.}
}
\value{
Logical, \code{TRUE} if internet site exists, \code{FALSE} otherwise

Logical, \code{TRUE} if internet site exists, \code{FALSE} otherwise.
}
\description{
A lightweight function that may be less reliable than more purpose built solutions
such as checking a specific web page using \code{RCurl::url.exists}. However, this is
slightly faster and is sufficient for many uses.
}
