% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{studyAreaName}
\alias{studyAreaName}
\alias{studyAreaName,character-method}
\alias{studyAreaName,ANY-method}
\title{Get a unique name for a given study area}
\usage{
studyAreaName(studyArea, ...)

\S4method{studyAreaName}{character}(studyArea, ...)

\S4method{studyAreaName}{ANY}(studyArea, ...)
}
\arguments{
\item{studyArea}{Spatial object.}

\item{...}{Other arguments (not currently used)}
}
\value{
A character string using the \code{.robustDigest} of the \code{studyArea}. This is only intended
for use with spatial objects.
}
\description{
Digest a spatial object to get a unique character string (hash) of the study area.
Use \code{.suffix()} to append the hash to a filename,
e.g., when using \code{filename2} in \code{prepInputs}.
}
\examples{
studyAreaName("Ontario")
}
