% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/robustDigest.R
\name{.removeCacheAtts}
\alias{.removeCacheAtts}
\title{Remove attributes that are highly varying}
\usage{
.removeCacheAtts(x)
}
\arguments{
\item{x}{Any arbitrary R object that could have attributes}
}
\description{
Remove attributes that are highly varying
}
