% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcessTo.R
\name{.isGridded}
\alias{.isGridded}
\alias{isSpatial}
\alias{.isVector}
\alias{.isSF}
\alias{.isSpat}
\alias{.isSpatialAny}
\alias{.isCRSany}
\title{Some spatial helper functions}
\usage{
.isGridded(x)

.isVector(x)

.isSF(x)

.isSpat(x)

.isSpatialAny(x)

.isCRSany(x)
}
\arguments{
\item{x}{A spatial object.}
}
\value{
Logical.
}
\description{
Some spatial helper functions
}
\details{
\code{.isGridded} returns \code{TRUE} if the object is a \code{SpatRaster} or \code{Raster}

\code{.isVector} returns \code{TRUE} if the object is \code{SpatVector}, \code{spatial} or \code{sf}

\code{.isSF} returns \code{TRUE} if the object is \code{sf} or \code{sfc}

\code{.isSpat} returns \code{TRUE} if the object is \code{SpatVector} or \code{SpatRaster}

\code{.isSpatialAny} returns \code{TRUE} if the object returns \code{TRUE} for \code{.isGridded} or
\code{.isVector}
}
