% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{.formalsNotInCurrentDots}
\alias{.formalsNotInCurrentDots}
\title{Identify which formals to a function are not in the current \code{...}}
\usage{
.formalsNotInCurrentDots(fun, ..., dots, formalNames, signature = character())
}
\arguments{
\item{fun}{A function}

\item{...}{The ... from inside a function. Will be ignored if \code{dots} is
provided explicitly.}

\item{dots}{Optional. If this is provided via say \code{dots = list(...)},
then this will cause the \code{...} to be ignored.}

\item{formalNames}{Optional character vector. If provided then it will override the \code{fun}}
}
\value{
A list of the formals of the \code{fun} that are missing from the \code{...} or \code{dots}.
}
\description{
Advanced use.
}
\keyword{internal}
