% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{basename2}
\alias{basename2}
\title{A version of \code{base::basename} that is \code{NULL} resistant}
\usage{
basename2(x)
}
\arguments{
\item{x}{A character vector of paths}
}
\value{
\code{NULL} if x is \code{NULL}, otherwise, as \code{basename}.

Same as \code{\link[base:basename]{base::basename()}}
}
\description{
A version of \code{base::basename} that is \code{NULL} resistant
}
