% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checksums.R
\name{.digest}
\alias{.digest}
\alias{.digest,character-method}
\title{Calculate the hashes of multiple files}
\usage{
.digest(file, quickCheck = FALSE, ...)

\S4method{.digest}{character}(file, quickCheck = FALSE, algo = "xxhash64", ...)
}
\arguments{
\item{file}{Character vector of file paths.}

\item{quickCheck}{Logical indicating whether to use a fast file size check as a heuristic
for determining changes to a file.}

\item{...}{Additional arguments to \code{\link[digest:digest]{digest::digest()}}.}
}
\value{
A character vector of hashes.
}
\description{
Internal function. Wrapper for \code{\link[digest:digest]{digest::digest()}} using \code{algo = xxhash64}.
}
\author{
Alex Chubaty
}
\keyword{internal}
